/**
 * @file clipboard.cc
 * @brief åץܡɼν
 *
 * @author Yashiro Takeshi <yashiromann@nifty.com> 
 * @date 2008-05-01
 *
 * Copyright (C) 2008, Yashiro Takeshi
 */

#include <basic.h>
#include <bstdio.h>
#include <tcode.h>
#include <tstring.h>
#include <btron/tf.h>
#include <btron/libapp.h>
#include <btron/taddata.h>
#include <tad.h>
#include <libimg.h>
#include <vector>
#include <map>
#include <string>
#include "int.h"
#include "synergy.h"

extern WERR so_writeall(W s, B* buf, W len);


#define BMP_NMAXPLANE 4

struct BMP_EXT {
  UW   planes;
  UH   pixbits;
  UH   rowbytes;
  RECT bounds;
  UB*  baseaddr[BMP_NMAXPLANE];
};


struct SynergyBmpIoSource {
  UINT8* buf;
  int len; // bufʬĹ
  int pos;
};


ERR synergy_bmp_readfn(UB* bytes, UW reqsize, VOID* io_src)
{
  SynergyBmpIoSource* io = (SynergyBmpIoSource*)io_src;

  int i = 0;

  if (io->pos < 14) {
    UB bmpFileHeader[14] = {
      'B', 'M', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
    };
#if BIGENDIAN
    UINT32 bfSize = swap_int32(14 + io->len);
    UINT32 bfOffBits = swap_int32(14 + 40);
#else
    UINT32 bfSize = 14 + io->len;
    UINT32 bfOffBits = 14 + 40;
#endif
    memcpy((B*)&(bmpFileHeader[2]), (B*)&bfSize, sizeof(UINT32));
    memcpy((B*)&(bmpFileHeader[10]), (B*)&bfOffBits, sizeof(UINT32));

    if (reqsize <= 14 - io->pos) {
      memcpy((B*)bytes, (B*)&(bmpFileHeader[io->pos]), reqsize);
      io->pos += reqsize;
      return reqsize;
    }

    memcpy((B*)bytes, (B*)&(bmpFileHeader[io->pos]), 14 - io->pos);
    i += 14 - io->pos;
    io->pos = 14;
  }

  int rlen = reqsize - i;
  if (rlen > io->len - (io->pos - 14)) {
    rlen = io->len - (io->pos - 14);
  }
  if (rlen <= 0) {
    return i;
  }
  memcpy((B*)&(bytes[i]), (B*)&(io->buf[io->pos - 14]), rlen);
  io->pos += rlen;

  return i + rlen;
}


ERR synergy_bmp_writefn(VOID* io_src, UB* bytes, UW size)
{
  SynergyBmpIoSource* io = (SynergyBmpIoSource*)io_src;

  int i = 0;

  if (io->pos < 14) {
    if (size < 14 - io->pos) {
      io->pos += size;
      return size;
    }

    i += 14 - io->pos;
    io->pos = 14;
  }

  int wlen = size - i;
  memcpy(&(io->buf[io->pos - 14]), (B*)&(bytes[i]), wlen);
  io->pos += wlen;

  return i + wlen;
}


ERR synergy_bmp_flushfn(VOID* io_src)
{
  return 0;
}


W synergy_put_clipboard(SynergyTaskInfo* info, UINT8* data, int len, 
                        int id)
{
  std::vector<TRAYREC> trec;
  TRAYREC tmp;
  TC name[13] = {TK_S, TK_Y, TK_N, TK_SHRP, TNULL, TNULL, TNULL, TNULL};
  name[4] = TK_0 + (id / 100) % 10;
  name[5] = TK_0 + (id / 10) % 10;
  name[6] = TK_0 + id % 10;
  static const TEXTSEG textseg = {
    {{0, 0, 0, 0}},
    {{0, 0, 0, 0}},
    -120, -120, 0x21, 0
  };
  static const TT_RATIO ratioseg = {
#if BIGENDIAN
    TT_RATIO_SID, 0,
#else
    0, TT_RATIO_SID,
#endif
    0, 0,
  };
  static const FIGSEG figseg = {
    {{0, 0, 0, 0}}, // rect
    {{0, 0, 0, 0}}, // rect
    -120, -120, 0
  };
  static const IMAGESEG imageseg = {
    {{0, 0, 0, 0}}, // rect
    {{0, 0, 0, 0}}, // rect
    -120, -120, 
    0, 0x0001,
    {0x1008, 0x0808, 0x0008, 0}, 
    0, 0, 0,
    // ̤ΰ̥ӥåȥޥå
    0, 1, 0, 0, {{0, 0, 0, 0}}, {0}
  };
  int inpos = 0;
  bool str_first = true;
  W ret = 0;

  if (len < sizeof(UINT32)) {
    return -1;
  }

  // եޥåȿ
  UINT32 nformats;
  memcpy((B*)&nformats, (B*)&(data[inpos]), sizeof(UINT32));
  nformats = ntoh_int32(nformats);
  inpos += sizeof(UINT32);

  // ʸϳϥȤ򥳥ԡ
  tmp.id  = TS_TEXT;
  tmp.len = sizeof(TEXTSEG);
  tmp.dt  = (B*)malloc(sizeof(TEXTSEG));
  memcpy(tmp.dt, (const B*)&textseg, sizeof(TEXTSEG));
  trec.push_back(tmp);

  // ʸ/̾ꥻ()򥳥ԡ
  tmp.id  = TS_TFONT;
  tmp.len = 6;
  tmp.dt  = (B*)malloc(sizeof(TT_RATIO));
  memcpy(tmp.dt, (const B*)&ratioseg, sizeof(TT_RATIO));
  trec.push_back(tmp);

  // ƥեޥåȤϡǼ
  for (int i = 0; i < nformats; i++) {
    UINT32 formtype, formsize;
    B *buf;
    int blen;
    bool cont = true;

    if (len - inpos < sizeof(UINT32) * 2) {
      ret = -1;
      goto err_ret;
    }

    memcpy((B*)&formtype, (B*)&(data[inpos]), sizeof(UINT32));
    formtype = ntoh_int32(formtype);
    inpos += sizeof(UINT32);
    memcpy((B*)&formsize, (B*)&(data[inpos]), sizeof(UINT32));
    formsize = ntoh_int32(formsize);
    inpos += sizeof(UINT32);

    if (len - inpos < formsize) {
      ret = -1;
      goto err_ret;
    }

    switch (formtype) {
    case 0: // kText
      {
        W dlen = 0;
        TC *str;

        buf = (B*)&(data[inpos]);
        blen = formsize;

        tf_strtotcs(info->tctx_from, (UB*)buf, blen,
                    (str_first ? TF_ATTR_START : 0) | TF_ATTR_HAN, 
                    NULL, &dlen);
        dlen++; // TNULLʬ䤷Ƥ
        str = (TC*)malloc(sizeof(TC) * dlen);
        tf_strtotcs(info->tctx_from, (UB*)buf, blen,
                    (str_first ? TF_ATTR_START : 0) | TF_ATTR_HAN, 
                    str, &dlen);

        tmp.id  = TR_TEXT;
        tmp.len = dlen * sizeof(TC);
        tmp.dt  = (B*)str;
        trec.push_back(tmp);

        str_first = false;
      }
      cont = false;
      break;
    case 1: // kBitmap
      {
        CSPEC cspec = {1, {0x1008, 0x808, 0x0008, 0}, NULL};
        BMP bmap;
        IMG_COMPACT compact = {LIBIMG_METHOD_BMP, LIBIMG_BMP_OPT_DEFAULT};
        IMG_BMP img_bmp;
        SynergyBmpIoSource io = {(UB*)&(data[inpos]), formsize, 0};
        IMAGESEG* imgseg;
        FIGSEG* fseg;

        bmap.pixbits = 0x2018;
        img_bmp.color_spec = &cspec;
        img_bmp.bmap = &bmap;
        img_bmp.funcptr = (FUNCP)synergy_bmp_readfn;
        img_bmp.flushptr = NULL;
        img_bmp.io_src = &io;

        W err = libimg_rea_bmp(&img_bmp, &compact);
        if (err < 0) {
          break;
        }

        // (1) ޷ϥ
        fseg = (FIGSEG*)malloc(sizeof(FIGSEG));
        memcpy((B*)fseg, (const B*)&figseg, sizeof(FIGSEG));
        fseg->view = bmap.bounds;
        fseg->draw = bmap.bounds;

        tmp.id  = TS_FIG;
        tmp.len = sizeof(FIGSEG);
        tmp.dt  = (B*)fseg;
        trec.push_back(tmp);

        // (2) 
        imgseg = (IMAGESEG*)malloc
          (sizeof(IMAGESEG)
           + sizeof(UW)
           + bmap.rowbytes * (bmap.bounds.c.bottom - bmap.bounds.c.top));
        memcpy((B*)imgseg, (const B*)&imageseg, sizeof(IMAGESEG));
        imgseg->view = bmap.bounds;
        imgseg->draw = bmap.bounds;
        imgseg->pixbits = bmap.pixbits;
        imgseg->rowbytes = bmap.rowbytes;
        imgseg->bounds = bmap.bounds;
        imgseg->base_off[0] = BASE_SIZE_IMAGESEG + sizeof(UW);
        B* imgseg_b = (B*)imgseg;
        memcpy(&(imgseg_b[BASE_SIZE_IMAGESEG + sizeof(UW)]), 
               bmap.baseaddr[0], 
               bmap.rowbytes * (bmap.bounds.c.bottom - bmap.bounds.c.top));

        tmp.id  = TS_IMAGE;
        tmp.len = BASE_SIZE_IMAGESEG + sizeof(UW) + 
          bmap.rowbytes * (bmap.bounds.c.bottom - bmap.bounds.c.top);
        tmp.dt  = (B*)imgseg;
        trec.push_back(tmp);

        // (3) ޷λ
        tmp.id  = TS_FIGEND;
        tmp.len = 0;
        tmp.dt  = NULL;
        trec.push_back(tmp);

        free(bmap.baseaddr[0]);
      }
      cont = false;
      break;
    default:
      break;
    }

    if (!cont) {
      break;
    }

    inpos += formsize;
  }

  // ȥ졼˳Ǽ
  wai_sem(info->lock, T_FOREVER);
  tpsh_dat(&(trec[0]), trec.size(), name);
  sig_sem(info->lock);

  ret = 0;

  // FALLTHROUGH
err_ret:
  int iend = trec.size();
  for (int i = 0; i < iend; i++) {
    free(trec[i].dt);
  }
  return ret;
}


B* synergy_append_imageseg(SynergyTaskInfo* info, 
                           IMAGESEG* seg, int len)
{
  CSPEC cspec_bmp;
  IMG_COMPACT compact = {LIBIMG_METHOD_BMP, LIBIMG_BMP_OPT_DEFAULT};
  IMG_BMP img_bmp;
  BMP bmap;
  UB* seg_b = reinterpret_cast<UB*>(seg);
  SynergyBmpIoSource io_src;
  B* buf = NULL;
  B* buf2 = NULL;
  int blen;
  int b2len;
  W err;

  if (seg->compac != 0 || seg->planes > BMP_NMAXPLANE) {
    return NULL;
  }

  cspec_bmp.attr = (seg->color & 0xf) | 0x2000;
  cspec_bmp.info[0] = seg->cinfo[0];
  cspec_bmp.info[1] = seg->cinfo[1];
  cspec_bmp.info[2] = seg->cinfo[2];
  cspec_bmp.info[3] = seg->cinfo[3];
  if ((seg->color & 0x8) != 0) {
    // ȥ륨ǥǤ⤳ν֤ɤߤ
    int cmap_offset = (seg->cinfo[2] << 16) | seg->cinfo[3];
    cspec_bmp.colmap = (COLOR*)&(seg_b[cmap_offset]);
  }
  else {
    cspec_bmp.colmap = NULL;
  }

  blen = 40 + 4 * 
    + (seg->bounds.c.right - seg->bounds.c.left)
    * (seg->bounds.c.bottom - seg->bounds.c.top);
  b2len = 4 * (seg->bounds.c.right - seg->bounds.c.left)
    * (seg->bounds.c.bottom - seg->bounds.c.top);

  if (seg->planes == 1 && seg->pixbits == 0x2018) {
    bmap.planes = seg->planes;
    bmap.pixbits = seg->pixbits;
    bmap.rowbytes = seg->rowbytes;
    bmap.bounds = seg->bounds;
    bmap.baseaddr[0] = &(seg_b[seg->base_off[0]]);

    img_bmp.color_spec = &cspec_bmp;
    img_bmp.bmap = &bmap;
    (FUNCP)img_bmp.funcptr = (FUNCP)synergy_bmp_writefn;
    (FUNCP)img_bmp.flushptr = (FUNCP)synergy_bmp_flushfn;
    img_bmp.io_src = &io_src;
  }
  else {
    CSPEC cspec_fc = {0x2001, {0x1008, 0x808, 0x0008, 0}, NULL};
    BMP_EXT bmap_src;
    bool error = false;

    buf2 = (B*)malloc(b2len);

    bmap_src.planes = seg->planes;
    bmap_src.pixbits = seg->pixbits;
    bmap_src.rowbytes = seg->rowbytes;
    bmap_src.bounds = seg->bounds;
    for (int i = 0; i < seg->planes; i++) {
      bmap_src.baseaddr[i] = &(seg_b[seg->base_off[i]]);
    }

    bmap.planes = 1;
    bmap.pixbits = 0x2018;
    bmap.rowbytes = 4 * (seg->bounds.c.right - seg->bounds.c.left);
    bmap.bounds = seg->bounds;
    bmap.baseaddr[0] = (UB*)buf2;

    W gid_src = -1, gid_dst = -1;
    gid_src = gopn_mem(NULL, (BMP*)&bmap_src, (B*)&cspec_bmp);
    if (gid_src < 0) {
      // 16ӥåȥޥåפξ硢planes = 4 ʤΤǡǼ(E_NOSPT)
      error = true;
      goto err_ret_colormapped;
    }
    gid_dst = gopn_mem(NULL, &bmap, (B*)&cspec_fc);
    if (gid_dst < 0) {
      error = true;
      goto err_ret_colormapped;
    }

    err = gcop_bmp(gid_src, &(bmap_src.bounds), gid_dst, NULL, 
                   NULL, G_STORE|G_CVFORM|G_CVCOLOR);
    if (err < 0) {
      error = true;
      goto err_ret_colormapped;
    }

    // FALLTHROUGH
  err_ret_colormapped:
    if (gid_src >= 0) {
      gcls_env(gid_src);
    }
    if (gid_dst >= 0) {
      gcls_env(gid_dst);
    }

    if (error) {
      free(buf2);
      return NULL;
    }

    img_bmp.color_spec = &cspec_fc;
    img_bmp.bmap = &bmap;
    (FUNCP)img_bmp.funcptr = (FUNCP)synergy_bmp_writefn;
    (FUNCP)img_bmp.flushptr = (FUNCP)synergy_bmp_flushfn;
    img_bmp.io_src = &io_src;
  }

  buf = (B*)malloc(sizeof(UINT32)*2 + blen + 1);
  io_src.buf = (UINT8*)&(buf[sizeof(UINT32)*2]);
  io_src.len = blen + 1;
  io_src.pos = 0;

  err = libimg_wri_bmp(&img_bmp, &compact);
  free(buf2);
  if (err < 0) {
    free(buf);
    return NULL;
  }

  UINT32* formtype = (UINT32*)buf;
  UINT32* formsize = (UINT32*)&(buf[sizeof(UINT32)]);

  *formtype = hton_int32(1); // kBitmap
  *formsize = hton_int32(io_src.pos - 14);

  return buf;
}


B* synergy_append_text(SynergyTaskInfo* info, TC* text, int len, 
                       bool& str_first)
{
  B* buf = NULL;
  W dlen;

  tf_tcstostr(info->tctx_to, text, len, 0x21, 
              (str_first ? TF_ATTR_START : 0) | TF_ATTR_HAN,
              NULL, &dlen);
  dlen++; // \0 ʬ䤷Ƥ
  buf = (B*)malloc(sizeof(UINT32) * 2 + dlen);
  tf_tcstostr(info->tctx_to, text, len, 0x21,
              (str_first ? TF_ATTR_START : 0) | TF_ATTR_HAN, 
              (UB*)&(buf[sizeof(UINT32)*2]), &dlen);

  UINT32* formtype = (UINT32*)buf;
  UINT32* formsize = (UINT32*)&(buf[sizeof(UINT32)]);

  *formtype = hton_int32(0); // kText
  *formsize = hton_int32(dlen);

  str_first = false;

  return buf;
}


W synergy_grab_clipboard(SynergyTaskInfo* info)
{
  UINT8 header[4+4+1+4];
  UINT32 h_len, h_seq;
  W err;

  h_len = hton_int32(4+1+4);
  h_seq = hton_int32(info->seq);
  memcpy(&(header[0]), (B*)&h_len, 4);
  memcpy(&(header[4]), "CCLP", 4);
  header[8] = 0;
  memcpy(&(header[9]), (B*)&h_seq, 4);

  err = so_writeall(info->socket, (B*)header, 13);
  if (err < 13) {
    return -1;
  }

  info->cid_mine[0] = true;
  info->cid_upd[0]  = true;

  return 0;
}


W synergy_get_clipboard(SynergyTaskInfo* info)
{
  W err, size;
  W ret = 0;
  int len;

  wai_sem(info->lock, T_FOREVER);
  err = tpop_dat(NULL, 0, &size, -1, NULL);
  sig_sem(info->lock);
  if (err < 0) {
    return -1;
  }

  TRAYREC* data = NULL;
  W ndata;

  std::vector<B*> outbuf;
  bool str_first = true;

  UINT8 header[4+4+1+4+4+4];
  UINT32 h_len, h_seq, h_strsize, h_nformats;
  int iend;
  std::vector<TC> s;
  bool cont = true;

  data = (TRAYREC*)malloc(size+1);
  ndata = tpop_dat((TRAYREC*)data, size, &size, -1, NULL);
  if (ndata < 0) {
    goto err_ret;
  }

  for (int i = 0; i < ndata; ) {
    if ((data[i].id & 0x4000) != 0) {
      len = 0;
      for (int j = 0; j < ndata; j++) {
        len += data[i].len;
        if ((data[i].id & 0x4000) != 0) {
          break;
        }
      }
      data[i].len = len;
    }

    switch (data[i].id & ~0x4000) {
    case TR_TEXT:
    case TR_FIG:
      {
        B* buf = data[i].dt;

        for (int j = 0; j < data[i].len; ) {
          TADSEG* tseg = (TADSEG*)&(buf[j]);
          LTADSEG* ltseg = (LTADSEG*)&(buf[j]);

          if (tseg->id >= 0xff80 && tseg->id <= 0xfffe) {
            B*  sbuf;
            int slen, jlen;

            if (tseg->len == 0xffff) {
              sbuf = &(buf[j + sizeof(LTADSEG)]);
              slen = ltseg->llen;
              jlen = slen + sizeof(LTADSEG);
            }
            else {
              sbuf = &(buf[j + sizeof(TADSEG)]);
              slen = tseg->len;
              jlen = slen + sizeof(TADSEG);
            }

            if ((tseg->id & 0xff) == TS_IMAGE && s.empty()) {
              B* out = synergy_append_imageseg(info, (IMAGESEG*)sbuf, slen);
              outbuf.push_back(out);
              cont = false;
            }
            j += jlen;
          }
          else if (tseg->id != 0) {
            s.push_back(tseg->id);
            j += sizeof(TC);
          }
        }
      }
      break;
    case TS_IMAGE:
      {
        if (s.empty()) {
          B* out = synergy_append_imageseg(info, (IMAGESEG*)data[i].dt, 
                                           data[i].len);
          outbuf.push_back(out);
          cont = false;
        }
      }
      break;
    default:
      break;
    }

    // ³ե饰ΰޤǥå
    while (i < ndata && (data[i].id & 0x4000) != 0) {
      i++;
    }
    i++;

    if (!cont) {
      break;
    }
  }

  if (!s.empty()) {
    B* out = synergy_append_text(info, &(s[0]), s.size(), str_first);
    outbuf.push_back(out);
    s.clear();
  }

  h_len = sizeof(header) - sizeof(UINT32);
  h_strsize = sizeof(UINT32);
  h_nformats = 0;
  iend = outbuf.size();
  for (int i = 0; i < iend; i++) {
    if (outbuf[i] != NULL) {
      B* out = outbuf[i];
      UINT32 olen;
      memcpy((B*)&olen, &(out[sizeof(UINT32)]), sizeof(UINT32));
      olen = ntoh_int32(olen);

      h_nformats++;
      h_len += sizeof(UINT32) * 2 + olen;
      h_strsize += sizeof(UINT32) * 2 + olen;
    }
  }
  h_len = hton_int32(h_len);
  h_strsize = hton_int32(h_strsize);
  h_nformats = hton_int32(h_nformats);
  h_seq = hton_int32(info->seq);

  memcpy(&(header[0]), (B*)&h_len, 4);
  memcpy(&(header[4]), "DCLP", 4);
  header[8] = 0;
  memcpy(&(header[9]), (B*)&h_seq, 4);
  memcpy(&(header[13]), (B*)&h_strsize, 4);
  memcpy(&(header[17]), (B*)&h_nformats, 4);

  err = so_writeall(info->socket, (B*)header, 21);
  if (err < 21) {
    goto err_ret2;
  }

  for (int i = 0; i < iend; i++) {
    if (outbuf[i] != NULL) {
      B* out = outbuf[i];
      UINT32 olen;
      memcpy((B*)&olen, &(out[sizeof(UINT32)]), sizeof(UINT32));
      olen = ntoh_int32(olen);

      err = so_writeall(info->socket, (B*)out, olen + sizeof(UINT32)*2);
      if (err < olen) {
        goto err_ret2;
      }
    }
  }

  ret = 0;
  goto err_ret1;

err_ret2:
  wai_sem(info->lock, T_FOREVER);
  if (info->socket >= 0) {
    so_close(info->socket);
    info->socket = -1;
  }
  sig_sem(info->lock);

err_ret1:
  iend = outbuf.size();
  for (int i = 0; i < iend; i++) {
    free(outbuf[i]);
  }

  // FALLTHROUGH
err_ret:
  free(data);
  return ret;
}

