/**
 * @file synergy.cc
 * @brief Synergy()
 *
 * @author Yashiro Takeshi <yashiromann@nifty.com> 
 * @date 2008-05-01
 *
 * Copyright (C) 2008, Yashiro Takeshi
 */

#include <basic.h>
#include <bstdio.h>
#include <bstring.h>
#include <bstdlib.h>
#include <tcode.h>
#include <tstring.h>
#include <keycode.h>
#include <btron/event.h>
#include <btron/proctask.h>
#include <btron/taskcomm.h>
#include <btron/bsocket.h>
#include <btron/libapp.h>
#include <btron/device.h>
#include <itron/itron.h>
#include <driver/kbpd.h>
#include "int.h"
#include "keysym.h"
#include "synergy.h"
#include "tcsliteral.h"

#define SYNERGY_PORT 24800
#define KPROTOCOL_VER_MAJOR 1
#define KPROTOCOL_VER_MINOR 3

extern DEV_SPEC SCREEN;

extern void dispError(const TC* errmsg);

extern int  keyid_to_physical(UINT16 keyid);

extern void synergy_set_pdabs(SynergyTaskInfo* info, PNT pos, bool enter);
extern void synergy_set_pdrel(SynergyTaskInfo* info, PNT pos);
extern void synergy_set_wheel(SynergyTaskInfo* info, W wheel);
extern void synergy_set_pdinput(SynergyTaskInfo* info, int btn, bool press);
extern void synergy_set_pdleave(SynergyTaskInfo* info);
extern void synergy_set_keyinput(SynergyTaskInfo* info, int key, bool press);

extern W synergy_put_clipboard(SynergyTaskInfo* info, UINT8* data, int len,
                               int id);
extern W synergy_get_clipboard(SynergyTaskInfo* info);
extern W synergy_grab_clipboard(SynergyTaskInfo* info);

void synergy_req_stop();



/**
 * åȤǡɤ߹.
 * åȤνüãϡso_close 
 */
WERR so_readall(SynergyTaskInfo* info, W s, B* buf, W len, bool* closed)
{
  W err;
  W pos = 0;
  fd_set rfds;
  TIMEVAL tmout;
  tmout.tv_sec = 0;
  tmout.tv_usec = 500000;

  while (pos < len) {
    // ȥ졼ιåѥåȼԤƱ˹Ԥ
    W  tnum, size;
    UW tsize;
    TC name[13];

    err = tpop_dat(NULL, 0, &size, -1, name);
    if (err <= 0) {
      goto label_select;
    }
    if (name[0] == TK_S && name[1] == TK_Y 
        && name[2] == TK_N && name[3] == TK_SHRP) {
      // Фäåץܡɤ֤ʤ
      goto label_select;
    }
    tnum = tget_sts(&tsize, NULL);
    if (tsize == info->tsize && tnum == info->tnum) {
      goto label_select;
    }
    info->tsize = tsize;
    info->tnum  = tnum;

    synergy_grab_clipboard(info);

  label_select:
    FD_ZERO(&rfds);
    FD_SET(s, &rfds);
    err = so_select(s+1, &rfds, NULL, NULL, &tmout);
    if (err == 0) {
      continue;
    }

    // ѥåȤ
    err = so_read(s, &(buf[pos]), len - pos);
    if (err < 0) {
      if (closed != NULL) {
        *closed = true;
      }
      so_close(s);
      return err;
    }
    else if (err == 0) {
      if (closed != NULL) {
        *closed = true;
      }
      so_close(s);
      return pos;
    }
    pos += err;
  }

  if (closed != NULL) {
    *closed = false;
  }
  return len;
}


/**
 * åȤؤǡ񤭹
 */
WERR so_writeall(W s, B* buf, W len)
{
  W err;
  W pos = 0;

  while (pos < len) {
    err = so_write(s, &(buf[pos]), len - pos);
    if (err < 0) {
      return err;
    }
    pos += err;
  }

  return len;
}


/**
 * ۥ̾̾
 */
sockaddr_in* sock_resolve(char* name, unsigned short port)
{
  W err;
  sockaddr_in* saddr = NULL;
  char tmp[HBUFLEN];
  char addr[5];
  hostent hostip;

  if (inet_pton(AF_INET, name, addr) <= 0) {
    err = so_gethostbyname(name, &hostip, tmp);
    if (err < 0) {
      return NULL;
    }
    memcpy((B*)addr, (B*)hostip.h_addr, hostip.h_length);
  }

  saddr = new sockaddr_in;
  memset((B*)saddr, sizeof(sockaddr_in), 0);
  saddr->sin_family = PF_INET;
  saddr->sin_port = hton_int16(port);
  memcpy((B*)&(saddr->sin_addr), (B*)addr, 4);

  return saddr;
}


/**
 * kMsgHelloХå
 */
void synergy_exec_kMsgHello(SynergyTaskInfo* info, UINT8* data, int len)
{
  W err;
  int pos_in = 7;
  int namelen = strlen(info->name);
  UINT8 buf[128];
  int pos_out = 0;
  const char* errmsg_reply = 
    "Failed to send kMsgHelloBack. Aborting.\n";

  struct SynProtoHello {
    UINT16 ver_major;
    UINT16 ver_minor;
  } hello;

  struct SynProtoHelloBack {
    UINT32 len;
    UINT8  cmd[7];
    // no padding here...
    UINT16 ver_major;
    UINT16 ver_minor;
    UINT32 namelen;
    // screen name follows...
  } hello_back;

  if (len - pos_in < sizeof(hello)) {
    printf("Synergy kMsgHello too short. Ignoring.\n");
    return;
  }

  hello.ver_major = 
    ntoh_int16(*reinterpret_cast<UINT16*>(&data[pos_in]));
  pos_in += sizeof(UINT16);
  hello.ver_minor = 
    ntoh_int16(*reinterpret_cast<UINT16*>(&data[pos_in]));
  pos_in += sizeof(UINT16);

  // ȤϡŬ kMsgHelloBack ֤
  hello_back.len = hton_int32(7 + sizeof(UINT16)*2 
                              + sizeof(UINT32) + namelen);
  memcpy((B*)hello_back.cmd, "Synergy", 7);
  hello_back.ver_major = hton_int16(KPROTOCOL_VER_MAJOR);
  hello_back.ver_minor = hton_int16(KPROTOCOL_VER_MINOR);
  hello_back.namelen = hton_int32(namelen);

  memcpy((B*)&buf[pos_out], (B*)&hello_back, sizeof(UINT32) + 7);
  pos_out += sizeof(UINT32) + 7;
  memcpy((B*)&buf[pos_out], (B*)&(hello_back.ver_major),
         sizeof(UINT16) * 2 + sizeof(UINT32));
  pos_out += sizeof(UINT16) * 2 + sizeof(UINT32);
  if (namelen <= sizeof(buf) - pos_out) {
    memcpy((B*)&buf[pos_out], info->name, namelen);
    pos_out += namelen;
    err = so_writeall(info->socket, (B*)buf, pos_out);
    if (err < pos_out) {
      printf(errmsg_reply);
      goto err_ret;
    }
  }
  else {
    err = so_writeall(info->socket, (B*)buf, pos_out);
    if (err < pos_out) {
      printf(errmsg_reply);
      goto err_ret;
    }
    err = so_writeall(info->socket, (B*)info->name, namelen);
    if (err < namelen) {
      printf(errmsg_reply);
      goto err_ret;
    }
  }

  for (int i = 0; i < 2; i++) {
    info->cid_mine[i] = false;
    info->cid_upd[i]  = false;
  }

  info->handshaked = true;
  return;

err_ret:
  wai_sem(info->lock, T_FOREVER);
  if (info->socket >= 0) {
    so_close(info->socket);
    info->socket = -1;
  }
  sig_sem(info->lock);
}


/**
 * kMsgDMouseMoveХå
 */
void synergy_exec_kMsgDMouseMove(SynergyTaskInfo* info, UINT8* data, int len)
{
  struct SynProtoMouseMove {
    UINT8 cmd[4];
    INT16 x;
    INT16 y;
  };

  PNT pos;
  SynProtoMouseMove* mouse_move = 
    reinterpret_cast<SynProtoMouseMove*>(data);

  if (len < sizeof(SynProtoMouseMove)) {
    printf("Insufficient argument for command kMsgDMouseMove. Ignoring.\n");
    return;
  }

  pos.x = ntoh_int16(mouse_move->x);
  pos.y = ntoh_int16(mouse_move->y);

  synergy_set_pdabs(info, pos, false);
}


/**
 * kMsgDRelMouseMoveХå
 */
void synergy_exec_kMsgDRelMouseMove(SynergyTaskInfo* info, 
                                    UINT8* data, int len)
{
  struct SynProtoRelMouseMove {
    UINT8 cmd[4];
    INT16 x;
    INT16 y;
  };

  PNT pos;
  SynProtoRelMouseMove* mouse_relmove = 
    reinterpret_cast<SynProtoRelMouseMove*>(data);

  if (len < sizeof(SynProtoRelMouseMove)) {
    printf("Insufficient argument for command kMsgDRelMouseMove. Ignoring.\n");
    return;
  }

  pos.x = ntoh_int16(mouse_relmove->x);
  pos.y = ntoh_int16(mouse_relmove->y);

  synergy_set_pdrel(info, pos);
}


/**
 * kMsgDKeyRepeatХå
 */
void synergy_exec_kMsgDKeyRepeat(SynergyTaskInfo* info, 
                                 UINT8* data, int len)
{
  struct SynProtoKeyRepeat {
    UINT8 cmd[4];
    UINT16 keyid;
    UINT16 modifier;
    // extra parameters follow
  };

  if (len < 4 + sizeof(UINT16)*2) {
    printf("Insufficient argument for command kMsgDKeyRepeat. Ignoring.\n");
    return;
  }

  // ԡȤΤΤϲ⤷ʤ
}


/**
 * kMsgCKeepAliveХå
 */
void synergy_exec_kMsgCKeepAlive(SynergyTaskInfo* info, UINT8* data, int len)
{
  struct SynProtoKeepAlive {
    UINT32 len;
    UINT8  cmd[4];
  };

  W err;
  SynProtoKeepAlive keep_alive;

  keep_alive.len = hton_int32(4);
  memcpy(keep_alive.cmd, "CALV", 4);

  err = so_writeall(info->socket, (B*)&keep_alive, 
                    sizeof(SynProtoKeepAlive));
  if (err < sizeof(SynProtoKeepAlive)) {
    printf("Failed to send kMsgCKeepAlive. Aborting.\n");
    goto err_ret;
  }

  return;

err_ret:
  wai_sem(info->lock, T_FOREVER);
  if (info->socket >= 0) {
    so_close(info->socket);
    info->socket = -1;
  }
  sig_sem(info->lock);  
}


/**
 * kMsgDMouseDownХå
 */
void synergy_exec_kMsgDMouseDown(SynergyTaskInfo* info, UINT8* data, int len)
{
  if (len < 5) {
    printf("Insufficient argument for command kMsgDMouseDown. Ignoring.\n");
    return;
  }

  if (data[4] == 1 || data[4] == 2 || data[4] == 3) {
    synergy_set_pdinput(info, data[4], true);
  }
  else {
    printf("Unknown button %d for KMsgDMouseDown.\n");
  }
}


/**
 * kMsgDMouseUpХå
 */
void synergy_exec_kMsgDMouseUp(SynergyTaskInfo* info, UINT8* data, int len)
{
  if (len < 5) {
    printf("Insufficient argument for command kMsgDMouseUp. Ignoring.\n");
    return;
  }

  if (data[4] == 1 || data[4] == 2 || data[4] == 3) {
    synergy_set_pdinput(info, data[4], false);
  }
  else {
    printf("Unknown button %d for KMsgDMouseUp.\n");
  }
}


/**
 * kMsgDMouseUpХå
 */
void synergy_exec_kMsgDMouseWheel(SynergyTaskInfo* info, 
                                  UINT8* data, int len)
{
  struct SynProtoMouseWheel {
    UINT8 cmd[4];
    INT16 delta1;
    INT16 delta2;
  };

  SynProtoMouseWheel* mwheel =
    reinterpret_cast<SynProtoMouseWheel*>(data);
  int ydelta;

  if (len < sizeof(INT16)) {
    printf("Insufficient argument for command kMsgDMouseWheel. Ignoring.\n");
    return;
  }
  else if (len < sizeof(SynProtoMouseWheel)) {
    ydelta = -(INT16)ntoh_int16(mwheel->delta1) / 120;
  }
  else {
    ydelta = -(INT16)ntoh_int16(mwheel->delta2) / 120;
  }

  synergy_set_wheel(info, ydelta);
}


/**
 * kMsgDKeyDownХå
 */
void synergy_exec_kMsgDKeyDown(SynergyTaskInfo* info, UINT8* data, int len)
{
  struct SynProtoKeyDown {
    UINT8  cmd[4];
    UINT16 keyid;
    UINT16 modifier;
    UINT16 keybtn;
  };

  SynProtoKeyDown* keydown =
    reinterpret_cast<SynProtoKeyDown*>(data);
  UINT16 btn;
  int index;

  if (len < 4 + sizeof(UINT16)*2) {
    printf("Insufficient argument for command kMsgDKeyDown. Ignoring.\n");
    return;
  }

  btn = ntoh_int16(keydown->keyid);
  index = keyid_to_physical(btn);
  if (index == 0) {
    printf("Unknown key down %04x\n", btn);
    return;
  }

  synergy_set_keyinput(info, index, true);
}


/**
 * kMsgDKeyUpХå
 */
void synergy_exec_kMsgDKeyUp(SynergyTaskInfo* info, UINT8* data, int len)
{
  struct SynProtoKeyUp {
    UINT8  cmd[4];
    UINT16 keyid;
    UINT16 modifier;
    UINT16 keybtn;
  };

  SynProtoKeyUp* keyup =
    reinterpret_cast<SynProtoKeyUp*>(data);
  UINT16 btn;
  int index;

  if (len < 4 + sizeof(UINT16)*2) {
    printf("Insufficient argument for command kMsgDKeyUp. Ignoring.\n");
    return;
  }

  btn = ntoh_int16(keyup->keyid);
  index = keyid_to_physical(btn);
  if (index == 0) {
    printf("Unknown key up %04x\n", btn);
    return;
  }

  synergy_set_keyinput(info, index, false);
}


/**
 * kMsgCClipboardХå
 */
void synergy_exec_kMsgCClipboard(SynergyTaskInfo* info, 
                                 UINT8* data, int len)
{
  if (data[4] < 2) {
    info->cid_mine[data[4]] = false;
  }
}


/**
 * kMsgDClipboardХå
 */
void synergy_exec_kMsgDClipboard(SynergyTaskInfo* info, 
                                 UINT8* data, int len)
{
  UINT32 cblen;
  UINT8* cbdata;

  if (len < 4+1+4+4) {
    printf("Insufficient argument for command kMsgDClipboard. Ignoring.\n");
    return;
  }

  if (data[4] >= 2) {
    return;
  }
  if (info->cid_upd[data[4]] && info->cid_mine[data[4]]) {
    return;
  }

  cbdata = &(data[4+1+4+4]);
  memcpy((B*)&cblen, (B*)&(data[4+1+4]), sizeof(UINT32));
  cblen = ntoh_int32(cblen);

  if (cblen > len - (4+1+4+4)) {
    printf("Illegal argument length for command kMsgDClipboard. Ignoring.\n");
    return;
  }    

  if (synergy_put_clipboard(info, cbdata, cblen, data[4]) < 0) {
    printf("Illegal marshalled data for command kMsgDClipboard. Ignoring.\n");
    return;
  }
}


/**
 * kMsgCEnterХå
 */
void synergy_exec_kMsgCEnter(SynergyTaskInfo* info, UINT8* data, int len)
{
  struct SynProtoCEnter {
    UINT8  cmd[4];
    UINT16 x;
    UINT16 y;
    UINT32 seq;
    UINT16 modifier;
  };

  SynProtoCEnter* center = 
    reinterpret_cast<SynProtoCEnter*>(data);
  PNT pos;

  if (len < sizeof(UINT16)*3 + sizeof(UINT32)) {
    printf("Insufficient argument for command kMsgCEnter. Ignoring.\n");
    return;
  }

  info->seq = ntoh_int32(center->seq);

  pos.x = ntoh_int16(center->x);
  pos.y = ntoh_int16(center->y);
  synergy_set_pdabs(info, pos, true);
}


/**
 * kMsgCLeaveХå
 */
void synergy_exec_kMsgCLeave(SynergyTaskInfo* info, UINT8* data, int len)
{
  W err, size, tnum;
  UW tsize;
  TC name[13];

  synergy_set_pdleave(info);

  if (info->cid_upd[0] && info->cid_mine[0]) {
    synergy_get_clipboard(info);
    info->cid_upd[0]  = false;
    info->cid_mine[0] = false;
  }
}


/**
 * kMsgCScreenSaverХå
 */
void synergy_exec_kMsgCScreenSaver(SynergyTaskInfo* info, 
                                   UINT8* data, int len)
{
  if (len < 5) {
    printf("Synergy kMsgCScreenSaver too short. Ignoring.\n");
  }
#if 0
  else {
    printf("Screensaver on server turned %s.\n", 
           (data[4] ? "on" : "off"));
  }
#endif
}


/**
 * kMsgCInfoAckХå
 */
void synergy_exec_kMsgCInfoAck(SynergyTaskInfo* info, UINT8* data, int len)
{
  ; // ⤷ʤ
}


/**
 * kMsgQInfoХå
 */
void synergy_exec_kMsgQInfo(SynergyTaskInfo* info, UINT8* data, int len)
{
  W err;
  int blen = 4 + sizeof(UINT16) * 7;
  PNT pos;

  struct SynProtoDInfo {
    UINT32 len;
    UINT8  cmd[4];
    UINT16 left;
    UINT16 top;
    UINT16 width;
    UINT16 height;
    UINT16 warp;
    UINT16 x;
    UINT16 y;
    // no padding here
  } dinfo;

  get_pdp(&pos);
  
  dinfo.len = hton_int32(blen);
  memcpy(dinfo.cmd, "DINF", 4);
  dinfo.left = hton_int16(0);
  dinfo.top = hton_int16(0);
  dinfo.width = hton_int16(SCREEN.hpixels);
  dinfo.height = hton_int16(SCREEN.vpixels);
  dinfo.warp = hton_int16(0);
  dinfo.x = hton_int16(pos.x);
  dinfo.y = hton_int16(pos.y);

  err = so_writeall(info->socket, (B*)&dinfo, 
                    sizeof(UINT32) + blen);
  if (err < sizeof(UINT32) + blen) {
    printf("Failed to send kMsgDInfo. Aborting.\n");
    goto err_ret;
  }

  return;

err_ret:
  wai_sem(info->lock, T_FOREVER);
  if (info->socket >= 0) {
    so_close(info->socket);
    info->socket = -1;
  }
  sig_sem(info->lock);
}


/**
 * kMsgCResetOptionsХå
 */
void synergy_exec_kMsgCResetOptions(SynergyTaskInfo* info, 
                                    UINT8* data, int len)
{
  ; // ̤ʤΤ̵
}


/**
 * kMsgDSetOptionsХå
 */
void synergy_exec_kMsgDSetOptions(SynergyTaskInfo* info, 
                                  UINT8* data, int len)
{
  ; // ̤ʤΤ̵
}


/**
 * kMsgCNoopХå
 */
void synergy_exec_kMsgCNoop(SynergyTaskInfo* info, UINT8* data, int len)
{
  ; // ⤷ʤ
}


/**
 * kMsgCCloseХå
 */
void synergy_exec_kMsgCClose(SynergyTaskInfo* info, UINT8* data, int len)
{
  wai_sem(info->lock, T_FOREVER);
  if (info->socket >= 0) {
    so_close(info->socket);
    info->socket = -1;
  }
  sig_sem(info->lock);
}


/**
 * kMsgEIncompatibleХå
 */
void synergy_exec_kMsgEIncompatible(SynergyTaskInfo* info, 
                                    UINT8* data, int len)
{
  struct SynProtoIncompatible {
    UINT8 cmd[4];
    UINT16 ver_major;
    UINT16 ver_minor;
  };

  SynProtoIncompatible* errpkt = 
    reinterpret_cast<SynProtoIncompatible*>(data);

  if (len < sizeof(SynProtoIncompatible)) {
    printf("Synergy kEIncompatible too short. Ignoring.\n");
    goto err_ret;
  }

  printf("Server cannot handle synergy protocol version %d.%04x. "
         "Aborting.\n", 
         ntoh_int16(errpkt->ver_major), 
         ntoh_int16(errpkt->ver_minor));

err_ret:
  synergy_req_stop();
}


/**
 * kMsgEBusyХå
 */
void synergy_exec_kMsgEBusy(SynergyTaskInfo* info, UINT8* data, int len)
{
  printf("Screen name %s already registered on server. Aborting.\n",
         info->name);

  synergy_req_stop();
}


/**
 * kMsgEUnknownХå
 */
void synergy_exec_kMsgEUnknown(SynergyTaskInfo* info, UINT8* data, int len)
{
  printf("Screen name %s not registered on server. Aborting.\n",
         info->name);

  synergy_req_stop();
}


/**
 * kMsgEBadХå
 */
void synergy_exec_kMsgEBad(SynergyTaskInfo* info, UINT8* data, int len)
{
  printf("Illegal command sent to server. Continuing anyway.\n");
}


// hello, hellobackϴޤޤʤ
SynProtoMapping synergy_cmd_list_[] = {
  { "DMMV", synergy_exec_kMsgDMouseMove },
  { "DMRM", synergy_exec_kMsgDRelMouseMove },
  { "DKRP", synergy_exec_kMsgDKeyRepeat }, 
  { "CALV", synergy_exec_kMsgCKeepAlive },
  { "DMDN", synergy_exec_kMsgDMouseDown },
  { "DMUP", synergy_exec_kMsgDMouseUp },
  { "DMWM", synergy_exec_kMsgDMouseWheel }, 
  { "DKDN", synergy_exec_kMsgDKeyDown },
  { "DKUP", synergy_exec_kMsgDKeyUp },
  { "DCLP", synergy_exec_kMsgDClipboard }, 
  { "CCLP", synergy_exec_kMsgCClipboard },
  { "CINN", synergy_exec_kMsgCEnter },
  { "COUT", synergy_exec_kMsgCLeave },
  { "CSEC", synergy_exec_kMsgCScreenSaver },
  { "CIAK", synergy_exec_kMsgCInfoAck },
  { "QINF", synergy_exec_kMsgQInfo }, 
  { "CROP", synergy_exec_kMsgCResetOptions }, 
  { "DSOP", synergy_exec_kMsgDSetOptions }, 
  { "CNOP", synergy_exec_kMsgCNoop },
  { "CBYE", synergy_exec_kMsgCClose },
  { "EICV", synergy_exec_kMsgEIncompatible },
  { "EBSY", synergy_exec_kMsgEBusy },
  { "EUNK", synergy_exec_kMsgEUnknown },
  { "EBAD", synergy_exec_kMsgEBad },
  { NULL,   NULL },
};


/**
 * SynergyޥɤФ륳Хåؿθ
 */
SynergyCallback* synergy_lookup_cmd(const UINT8* cmd)
{
  for (int i = 0; synergy_cmd_list_[i].key != NULL; i++) {
    if (strncmp((B*)cmd, (B*)synergy_cmd_list_[i].key, 4) == 0) {
      return synergy_cmd_list_[i].val;
    }
  }
  return NULL;
}


/**
 * SynergyΥޥɤμ¹
 */
void synergy_exec_cmd(SynergyTaskInfo* info, UINT8* data, int len)
{
  if (len >= 7 && strncmp((char*)data, "Synergy", 7) == 0) {
    synergy_exec_kMsgHello(info, data, len);
  }
  else if (len >= 4) {
    SynergyCallback *fun = synergy_lookup_cmd(data);
    if (fun != NULL) {
      (*fun)(info, data, len);
    }
    else {
      printf("Unknown command %c%c%c%c ignored.\n",
             data[0], data[1], data[2], data[3]);
    }
  }
  else {
    printf("Received incomplete synergy command. Aborting.\n");
    goto err_ret;
  }

  return;

err_ret:
  wai_sem(info->lock, T_FOREVER);
  if (info->socket >= 0) {
    so_close(info->socket);
    info->socket = -1;
  }
  sig_sem(info->lock);
}


/**
 * SynergyΥޥɤμӼ¹
 */
void synergy_receive_cmd(SynergyTaskInfo* info)
{
  W err;
  W socket = info->socket;
  UINT32 len;
  bool closed = false;
  UINT8 buf[256];
  UINT8* pbuf = buf;
  W tnum;
  UW tsize;

  err = so_readall(info, socket, (B*)&len, sizeof(len), &closed);
  if (err < sizeof(len)) {
    wai_sem(info->lock, T_FOREVER);
    if (!closed) {
      so_close(socket);
    }
    info->socket = -1;
    sig_sem(info->lock);
    goto err_ret;
  }

  len = ntoh_int32(len);

  if (len > sizeof(buf)) {
    if (info->handshaked) {
      pbuf = new UINT8[len];
    }
    else {
      // Ĺơˤ̤ΥӥäݤʷϵʤΤ
      printf("Initial synergy command with length %d > %d. Aborting\n",
             len, sizeof(buf));
      wai_sem(info->lock, T_FOREVER);
      if (!closed) {
        so_close(socket);
      }
      info->socket = -1;
      sig_sem(info->lock);
      goto err_ret;
    }
  }
  if (len > 0 && closed) {
    goto err_ret;
  }

  err = so_readall(info, socket, (B*)pbuf, len, &closed);
  if (err < len) {
    wai_sem(info->lock, T_FOREVER);
    if (!closed) {
      so_close(socket);
    }
    info->socket = -1;
    sig_sem(info->lock);
    goto err_ret;
  }

  synergy_exec_cmd(info, pbuf, len);

  // FALLTHROUGH
err_ret:
  if (pbuf != buf) {
    delete [] pbuf;
  }

  if (closed) {
    info->socket = -1;
  }
}


/**
 * Synergy
 */
void synergy_task(W arg)
{
  SynergyTaskInfo* info 
    = reinterpret_cast<SynergyTaskInfo*>(arg);

  for (;;) {
    // åȤ³
    while (info->socket < 0) {
      W err, sock;

      sock = so_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
      if (sock >= 0) {
        err = so_connect(sock,
                         reinterpret_cast<sockaddr*>(info->saddr),
                         sizeof(sockaddr_in));
        if (err >= 0) {
          info->socket = sock;
          break;
        }
        else {
          so_close(sock);
        }
      }

      dly_tsk(5000); // ³ޤ 5[s] ԤäƤߤ
    }

    // ޥɤΤȤ
    while (info->socket >= 0) {
      synergy_receive_cmd(info);
    }
  }
}


/**
 * Synergyư
 */
SynergyTaskInfo* synergy_start(const char* server, 
                               const char* name)
{
  SynergyTaskInfo* info = NULL;
  TC devname[] = {TK_s, TK_y, TK_n, TK_d, TK_r, TK_v, TNULL};
  W err, error;

  info = new SynergyTaskInfo;
  memset(info, sizeof(SynergyTaskInfo), 0);
  info->lock = -1;
  info->socket = -1;
  info->driver = -1;
  info->task_proto = -1;
  info->task_alive = -1;
  info->tnum = -1;
  info->tsize = -1;

  info->lock = cre_sem(1, SEM_EXCL|DELEXIT);
  if (info->lock < 0) {
    dispError((TC[]){EMSG_CRESEM});
    goto err_ret;
  }

  info->driver = opn_dev(devname, D_WRITE, &error);
  if (info->driver < 0) {
    dispError((TC[]){EMSG_OPNDRV});
    goto err_ret;
  }

  info->server = strdup(server);
  info->name = strdup(name);

  err = tf_open_ctx(&(info->tctx_from));
  if (err < 0) {
    dispError((TC[]){EMSG_TFCTX});
    info->tctx_from = NULL;
    goto err_ret;
  }
  tf_set_profile(info->tctx_from, tf_to_id(TF_ID_PROFSET_CONVERTFROM, "UTF-8"));

  err = tf_open_ctx(&(info->tctx_to));
  if (err < 0) {
    dispError((TC[]){EMSG_TFCTX});
    info->tctx_to = NULL;
    goto err_ret;
  }
  tf_set_profile(info->tctx_to, tf_to_id(TF_ID_PROFSET_CONVERTTO, "UTF-8"));
  tf_set_options(info->tctx_to, 
                 tf_to_id(TF_ID_OPT_CONVERT, "CONV_KANA_HAN"), 1);

  // Ф̾
  info->saddr = sock_resolve(info->server, SYNERGY_PORT);
  if (info->saddr == NULL) {
    printf("Resolving name %s failed. Aborting.\n", info->server);
    dispError((TC[]){EMSG_RESOLVE});
    goto err_ret;
  }

  info->handshaked = false;
  info->entered = false;
  synergy_set_pdleave(info);

  info->task_proto = cre_tsk(synergy_task, 192, (W)info);
  if (info->task_proto < 0) {
    dispError((TC[]){EMSG_CRETSK});
    goto err_ret;
  }

  return info;

err_ret:
  if (info != NULL) {
    if (info->lock >= 0) {
      del_sem(info->lock);
    }
    if (info->driver >= 0) {
      cls_dev(info->driver, 0, &error);
    }
    if (info->server != NULL) {
      free(info->server);
    }
    if (info->name != NULL) {
      free(info->name);
    }
    if (info->tctx_from != NULL) {
      tf_close_ctx(info->tctx_from);
    }
    if (info->tctx_to != NULL) {
      tf_close_ctx(info->tctx_to);
    }
    if (info->saddr != NULL) {
      delete info->saddr;
    }
    delete info;
  }
  return NULL;
}


/**
 * Synergyλ
 */
void synergy_stop(SynergyTaskInfo* info)
{
  W error;

  wai_sem(info->lock, T_FOREVER);
  if (info->task_alive < 0) {
    ter_tsk(info->task_alive);
  }
  if (info->task_proto < 0) {
    ter_tsk(info->task_proto);
  }
  sig_sem(info->lock);

  if (info->lock >= 0) {
    del_sem(info->lock);
  }
  if (info->driver >= 0) {
    cls_dev(info->driver, 0, &error);
  }
  if (info->server != NULL) {
    free(info->server);
  }
  if (info->name != NULL) {
    free(info->name);
  }
  if (info->tctx_from != NULL) {
    tf_close_ctx(info->tctx_from);
  }
  if (info->tctx_to != NULL) {
    tf_close_ctx(info->tctx_to);
  }
  if (info->saddr != NULL) {
    delete info->saddr;
  }
  delete info;
}


/**
 * SynergyĤƤ餦
 */
void synergy_req_stop()
{
  struct SynergyMsgReq {
    W msg_type;
    W msg_size;
    W msg_body;
  };

  SynergyMsgReq msg;
  msg.msg_type = MS_TYPE7;
  msg.msg_size = sizeof(W);
  msg.msg_body = 1;

  snd_msg(0, (MESSAGE*)&msg, NOWAIT);
  slp_tsk(T_FOREVER);
}

